-- rebuild 2D

--[==[

instancing = shader
{
    name = "instancing",
    
    properties =
    {
        {"instancesBuffer", "buffer"},
        {"s_texture", "texture"}
    },
    
    pass =
    {
        shadingModel = "unlit",
        cullFace = "none", -- backgrounds don't need culling
        blendMode = "additive", -- no need for blending
        
        vertex =
        [[
            //#version 430
            //#include <codea/common.glsl>
            
            struct Instance
            {
                vec4 position;
            };
            
            BUFFER_RW(instances, Instance, BUFFER0);
            
            void materialVertex(inout MaterialVertexInputs material)
            {
                // Append instance position :D
                material.worldPosition.xy += instances[gl_InstanceID].position.xy;
                //material.color = vec3(1,0,0);
            }
        ]],
        
        material = 
        [[
            //#version 430
            //#include <codea/common.glsl>
            
            void material(inout MaterialInputs material) 
            {
        /*
                vec2 uv = getUV0();
                prepareMaterial(material);
                material.baseColor = vec4(1.0, 0.5, 0.3, 1.0);
        */
        
            //vec2 vel = instances[gl_InstanceID].position.zw;
            material.baseColor = texture(s_texture, getUV0());
            //material.baseColor.rb += vel;
            }            
        ]],
        
        shading = 
        [[
        vec3 surfaceShading(const MaterialInputs materialInputs,
        const ShadingData shadingData,
        const LightData lightData) 
        {
            // Number of visible shade transitions
            const float shades = 5.0;
            // Ambient intensity
            const float ambient = 0.1;
            
            float toon = max(ceil(lightData.NdotL * shades) / shades, ambient);
            
            // Shadowing and attenuation
            toon *= lightData.visibility * lightData.attenuation;
            
            // Color and intensity
            vec3 light = lightData.colorIntensity.rgb * lightData.colorIntensity.w;
            
            return shadingData.diffuseColor * light * toon;
        }            
        ]]
    }
}

updatePositions = shader.compute(
    [[
        #version 430
        #include <codea/common.glsl>
        //#include <codea/noise/noise.glsl>
        
        // 定义实例结构体
        struct Instance
        {
            vec4 position;
        };
        
        BUFFER_RW(instances, Instance, BUFFER0);
        
        uniform uint instanceCount;
        
        NUM_THREADS(256, 1, 1)
    
        float rand(vec2 co){
            return fract(sin(dot(co.xy ,vec2(12.9898,78.233))) * 43758.5453);
        }

        // 
        vec3 SimplexPerlin2D_Deriv( vec2 P )
        {
        //  https://github.com/BrianSharpe/Wombat/blob/master/SimplexPerlin2D_Deriv.glsl
        
        //  simplex math constants
        const float SKEWFACTOR = 0.36602540378443864676372317075294;            // 0.5*(sqrt(3.0)-1.0)
        const float UNSKEWFACTOR = 0.21132486540518711774542560974902;          // (3.0-sqrt(3.0))/6.0
        const float SIMPLEX_TRI_HEIGHT = 0.70710678118654752440084436210485;    // sqrt( 0.5 )	height of simplex triangle
        const vec3 SIMPLEX_POINTS = vec3( 1.0-UNSKEWFACTOR, -UNSKEWFACTOR, 1.0-2.0*UNSKEWFACTOR );  //  simplex triangle geo
        
        //  establish our grid cell.
        P *= SIMPLEX_TRI_HEIGHT;    // scale space so we can have an approx feature size of 1.0
        vec2 Pi = floor( P + dot( P, vec2( SKEWFACTOR ) ) );
        
        // calculate the hash
        vec4 Pt = vec4( Pi.xy, Pi.xy + 1.0 );
        Pt = Pt - floor(Pt * ( 1.0 / 71.0 )) * 71.0;
        Pt += vec2( 26.0, 161.0 ).xyxy;
        Pt *= Pt;
        Pt = Pt.xzxz * Pt.yyww;
        vec4 hash_x = fract( Pt * ( 1.0 / 951.135664 ) );
        vec4 hash_y = fract( Pt * ( 1.0 / 642.949883 ) );
        
        //  establish vectors to the 3 corners of our simplex triangle
        vec2 v0 = Pi - dot( Pi, vec2( UNSKEWFACTOR ) ) - P;
        vec4 v1pos_v1hash = (v0.x < v0.y) ? vec4(SIMPLEX_POINTS.xy, hash_x.y, hash_y.y) : vec4(SIMPLEX_POINTS.yx, hash_x.z, hash_y.z);
        vec4 v12 = vec4( v1pos_v1hash.xy, SIMPLEX_POINTS.zz ) + v0.xyxy;
        
        //  calculate the dotproduct of our 3 corner vectors with 3 random normalized vectors
        vec3 grad_x = vec3( hash_x.x, v1pos_v1hash.z, hash_x.w ) - 0.49999;
        vec3 grad_y = vec3( hash_y.x, v1pos_v1hash.w, hash_y.w ) - 0.49999;
        vec3 norm = inversesqrt( grad_x * grad_x + grad_y * grad_y );
        grad_x *= norm;
        grad_y *= norm;
        vec3 grad_results = grad_x * vec3( v0.x, v12.xz ) + grad_y * vec3( v0.y, v12.yw );
        
        //	evaluate the kernel
        vec3 m = vec3( v0.x, v12.xz ) * vec3( v0.x, v12.xz ) + vec3( v0.y, v12.yw ) * vec3( v0.y, v12.yw );
        m = max(0.5 - m, 0.0);
        vec3 m2 = m*m;
        vec3 m4 = m2*m2;
        
        //  calc the derivatives
        vec3 temp = 8.0 * m2 * m * grad_results;
        float xderiv = dot( temp, vec3( v0.x, v12.xz ) ) - dot( m4, grad_x );
        float yderiv = dot( temp, vec3( v0.y, v12.yw ) ) - dot( m4, grad_y );
        
        //  Normalization factor to scale the final result to a strict 1.0->-1.0 range
        //  http://briansharpe.wordpress.com/2012/01/13/simplex-noise/#comment-36
        const float FINAL_NORMALIZATION = 99.204334582718712976990005025589;
        
        //  sum and return all results as a vec3
        return vec3( dot( m4, grad_results ), xderiv, yderiv ) * FINAL_NORMALIZATION;
        }

        void main()
        {

        // Get the current instance to work on
        uint i = gl_GlobalInvocationID.x;
        
        // Prevent going over the end of the buffer
        if (i < instanceCount)
        {
            vec2 pos = instances[i].position.xy;
            vec2 vel = instances[i].position.zw;
            vec2 screenSize = u_viewRect.zw / 2.0;

            // Sample two curl-noise fields that scroll in different directions and rates
            vec2 curl1 = SimplexPerlin2D_Deriv(pos * 0.01 + vec2(u_time.x, 0) * 0.1).yz;
            curl1 = vec2(-curl1.y, curl1.x) * 0.5;
            
            vec2 curl2 = SimplexPerlin2D_Deriv(pos * 0.05 + vec2(0, u_time.x) * 0.2).yz;
            curl2 = vec2(-curl2.y, curl2.x) * 0.5;
            
            // Apply curl noise velocity field
            vel += (curl1 + curl2) * u_time.y;
            
            // Apply some damping
            vel -= vel * 0.8 * u_time.y;
        
            // Add velocity to position * delta time
            pos += vel * 30.0 * u_time.y;
            
            // Bounce instances to screen bounds
            if (pos.x < 0.0)
            {
                vel.x *= -2.0; pos.x = 0.0;
            } 
                else if (pos.x > screenSize.x)
            {
            vel.x *= -2.0; pos.x = screenSize.x;
                    }
            if (pos.y < 0.0)
            {
            vel.y *= -2.0; pos.y = 0.0;
                    }
                    else if (pos.y > screenSize.y)
            {
            vel.y *= -2.0; pos.y = screenSize.y;
                    }
            
            instances[i].position = vec4(pos, vel);
        } 
    }           
    ]]
)

-- Use this function to perform your initial setup
function setup()
    
    -- Create a reusable shader chunk for the instances buffer that looks like this in GLSL:
    -- struct Instances 
    -- {
    --    vec4 position; // (x, y, vx, vy) position and velocity
    -- };
    
    --[=[
    instancesChunk = shader.chunk()
    :buffer{"Instance", "instances", access = "readwrite"}
    :vec4("position")
    :done()
    
    -- Create the instanced unlit sprite shader
    instancing = shader.builder()
    :unlit()
    :add(instancesChunk) -- make instances buffer available to shader
    :cull("none"):blend("additive")
    :texture{"sprite", value = asset.builtin.Cargo_Bot.Codea_Icon}
    :vertex
    [[
    // Append instance position :D
    material.worldPosition.xy += instances[gl_InstanceID].position.xy;
    //material.color = vec3(1,0,0);
    ]]
    :material
    [[
    //vec2 vel = instances[gl_InstanceID].position.zw;
    material.baseColor = texture(sprite, getUV0());
    //material.baseColor.rb += vel;
    ]]  
    :build()

    -- Compute shader for updating the instances each frame
    updatePositions = shader.builder()
    :add(instancesChunk)
    -- secret (for now) GPU noise library
    :include("codea/noise/noise.glsl")
    :uint{"instanceCount", value = 0}
    :threads(256,1,1)    
    :compute
    [[
    // Get the current instance to work on
    uint i = gl_GlobalInvocationID.x;
    
    // Prevent going over the end of the buffer
    if (i < instanceCount)
    {
        vec2 pos = instances[i].position.xy;
        vec2 vel = instances[i].position.zw;
        vec2 screenSize = u_viewRect.zw / 2.0;
        
        // Sample two curl-noise fields that scroll in different directions and rates
        vec2 curl1 = SimplexPerlin2D_Deriv(pos * 0.01 + vec2(u_time.x, 0) * 0.1).yz;
        curl1 = vec2(-curl1.y, curl1.x) * 0.5;
        
        vec2 curl2 = SimplexPerlin2D_Deriv(pos * 0.05 + vec2(0, u_time.x) * 0.2).yz;
        curl2 = vec2(-curl2.y, curl2.x) * 0.5;
        
        // Apply curl noise velocity field
        vel += (curl1 + curl2) * u_time.y;
        
        // Apply some damping
        vel -= vel * 0.8 * u_time.y;
        
        // Add velocity to position * delta time
        pos += vel * 30.0 * u_time.y;
    
        // Bounce instances to screen bounds
        if (pos.x < 0.0)
        {
            vel.x *= -2.0; pos.x = 0.0;
        } 
        else if (pos.x > screenSize.x)
        {
            vel.x *= -2.0; pos.x = screenSize.x;
        }
    
        if (pos.y < 0.0)
        {
            vel.y *= -2.0; pos.y = 0.0;
        }
        else if (pos.y > screenSize.y)
        {
            vel.y *= -2.0; pos.y = screenSize.y;
        }
    
        instances[i].position = vec4(pos, vel);
    }        
    ]]
    :build()
    
    --]=]

    -- Create a buffer to match the shader buffer (syntax is a bit ugly right now)
    local layout = shader.bufferLayout
    {
        {shader.position, 4, shader.float} -- vec4
    }
    -- Create buffer to hold instance data
    instancesBuffer = shader.buffer(layout)
    
    -- Create initial instances by appending data to buffer
    for k = 1, 200000 do
        instancesBuffer:append(shader.position, vec4(math.random(0, WIDTH), math.random(0, HEIGHT), 
        0, 0)) --math.random() * 0.5 - 0.5, math.random() * 0.5 - 0.5))
    end
    
    updatePositions:setBuffer("instancesBuffer", instancesBuffer, shader.readwrite)
    updatePositions.instanceCount = instancesBuffer.size
    
    instancing:setBuffer("instancesBuffer", instancesBuffer, shader.readwrite)
    
    instancing.s_texture = image.read(asset.builtin.Cargo_Bot.Codea_Icon)
    
    --[[
    -- Make one tiny little quad for our sprites
    m = mesh()    
    m:addRect(0, 0, 3, 3)
    -- m.material = material(instancing)
    m.shader = instancing
    --]]

    ---[[
    m = mesh()
    
    local size = vec2(2, 2)
    
    m.positions =
    {
        vec3(-size.x/2, size.y/2, 0),
        vec3(size.x/2, size.y/2, 0),
        vec3(size.x/2, -size.y/2, 0),
        vec3(-size.x/2, -size.y/2, 0)
    }
    
    m.uvs =
    {
        vec2(0, 0),
        vec2(1, 0),
        vec2(1, 1),
        vec2(0, 1)
    }
    
    m.colors =
    {
        color(233, 10, 10),
        color(30, 242, 5),
        color(7, 35, 250),
        color(255)
    }
    
    m:addElement(3,2,1)
    m:addElement(4,3,1)
    
    m.shader = instancing
    instancing.s_texture = image.read(asset.builtin.Cargo_Bot.Codea_Icon)
    -- instancing.s_texture = image.read(asset.builtin.Planet_Cute.Character_Cat_Girl)
    --]]
end

-- This function gets called once every frame
function draw()
    -- This sets a dark background color 
    background(40, 40, 50)
    
    local wx = updatePositions:workgroupSize()
    updatePositions:dispatch(math.ceil(instancesBuffer.size / wx), 1, 1)
    
    m:draw(instancesBuffer.size)
end

--]==]